package de.unikoblenz.fgbks.base.utils.boundary.checker;

import static de.unikoblenz.fgbks.base.utils.boundary.BoundType.EXCLUSIVE;
import static de.unikoblenz.fgbks.base.utils.boundary.BoundType.INCLUSIVE;

import de.unikoblenz.fgbks.base.utils.boundary.AbstractGrowingBoundary;

/**
 * The subsumes checker implementation of {@link BoundaryCheckType#SUBSUMES} for {@link
 * AbstractGrowingBoundary}. <br> The check returns true, if the first boundary completely subsumes
 * (not equal) the other boundary.
 * <br>
 * E.g: <br>
 *
 * <ul>
 *   <li>b1 = "[0..20]"; b2 = "[4..19]" -{@literal >} true
 *   <li>b1 = "[4..19]"; b2 = "[1..19]" -{@literal >} false
 *   <li>b1 = "[0..20]"; b2 = "[4..22]" -{@literal >} false
 * </ul>
 */
public class CheckSubsumes extends AbstractBoundaryCheck<AbstractGrowingBoundary> {

  private static final CheckSubsumes instance = new CheckSubsumes();

  private CheckSubsumes() {
    super();
  }

  /**
   * Get the current instance of this class.
   *
   * @return the singleton instance
   */
  public static CheckSubsumes getInstance() {
    return instance;
  }

  @Override
  public BoundaryCheckType getType() {
    return BoundaryCheckType.SUBSUMES;
  }

  @Override
  public boolean check(AbstractGrowingBoundary b1, AbstractGrowingBoundary b2) {
    return checkSubsumes(b1, b2);
  }

  /**
   * see {@link CheckSubsumes}
   */
  public static boolean checkSubsumes(AbstractGrowingBoundary b1, AbstractGrowingBoundary b2) {
    // TODO: check raw types
    int compLo = b1.getLowerBound().compareTo(b2.getLowerBound());
    int compUp = b1.getUpperBound().compareTo(b2.getUpperBound());
    if (compLo < 0 && compUp > 0) {
      return true;
    }
    if (compLo > 0
        || compUp < 0
        || (compLo == 0
        && b1.getLowerBoundType() == EXCLUSIVE
        && b2.getLowerBoundType() == INCLUSIVE)
        || (compUp == 0
        && b1.getUpperBoundType() == EXCLUSIVE
        && b2.getUpperBoundType() == INCLUSIVE)) {
      return false;
    }
    if (compLo == 0 && b1.getLowerBoundType() != b2.getLowerBoundType()
        || compUp == 0 && b1.getUpperBoundType() != b2.getUpperBoundType()) {
      return true;
    } else {
      return (compLo + compUp) != 0;
    }
  }
}
