package de.unikoblenz.fgbks.base.utils.boundary.impl;

import static de.unikoblenz.fgbks.base.utils.boundary.BoundType.EXCLUSIVE;
import static de.unikoblenz.fgbks.base.utils.boundary.BoundType.INCLUSIVE;
import static java.time.format.DateTimeFormatter.ISO_LOCAL_DATE;
import static java.time.temporal.ChronoField.HOUR_OF_DAY;
import static java.time.temporal.ChronoField.MINUTE_OF_HOUR;
import static java.time.temporal.ChronoField.SECOND_OF_MINUTE;

import de.unikoblenz.fgbks.base.utils.boundary.AbstractGrowingBoundary;
import de.unikoblenz.fgbks.base.utils.boundary.BoundType;
import de.unikoblenz.fgbks.base.utils.boundary.BoundaryParseException;
import java.time.LocalDateTime;
import java.time.format.DateTimeFormatter;
import java.time.format.DateTimeFormatterBuilder;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import org.apache.commons.lang3.Validate;

/**
 * The date boundary is ether a single date or a INCLUSIVE range between to dates (or between
 * -INFINITY and +INFINITY). The date boundary is represented as {@link LocalDateTime}. The text
 * representation matches the pattern {@link DateBoundary#datePatternString} and looks like: <br>
 *
 * <ul>
 *   <li>{@code "date and time("yyyy-MM-ddThh:mm:ss")"} <br>
 *       exact date
 *   <li>{@code "< date and time("yyyy-MM-ddThh:mm:ss")"} <br>
 *       lower than the given date
 *   <li>{@code "> date and time("yyyy-MM-ddThh:mm:ss")"} <br>
 *       higher than the given date
 *   <li>{@code "[date and time("yyyy-MM-ddThh:mm:ss")..date and time("yyyy-MM-ddThh:mm:ss")]"} <br>
 *       between the two given dates {@link BoundType#INCLUSIVE}
 *   <li>{@code "" (Empty string)}<br>
 *       All dates.
 * </ul>
 */
public class DateBoundary extends AbstractGrowingBoundary<LocalDateTime> {

  public static final String dateGroupName = "datetime";
  public static final String datePatternString =
      "^date and time\\(\"(?<"
          + dateGroupName
          + ">(?:19|20)\\d{2}-(?:0[1-9]|1[0-2])-(?:0[1-9]|[12]\\d|3[01])T(?:[01]\\d|2[0-3]):[0-5]\\d:[0-5]\\d)\"\\)$";
  public static final Pattern datePattern = Pattern.compile(datePatternString);
  public static final DateTimeFormatter CAMUNDA_DATE_TIME_FORMAT;

  static {
    CAMUNDA_DATE_TIME_FORMAT =
        new DateTimeFormatterBuilder()
            .appendLiteral("date and time(\"")
            .parseCaseInsensitive()
            .append(ISO_LOCAL_DATE)
            .appendLiteral('T')
            .appendValue(HOUR_OF_DAY, 2)
            .appendLiteral(':')
            .appendValue(MINUTE_OF_HOUR, 2)
            .appendLiteral(':')
            .appendValue(SECOND_OF_MINUTE, 2)
            .appendLiteral("\")")
            .toFormatter();
  }

  /**
   * Parse the input text to a new Date boundary. See {@link DateBoundary} for the required
   * structure.
   *
   * @param text the text, which will be parsed to the new Dateoundary object
   */
  public DateBoundary(String text) {
    super(text, LocalDateTime.class);
  }

  private DateBoundary() {
    super(LocalDateTime.class);
  }

  @Override
  public AbstractGrowingBoundary getCopy(
      LocalDateTime lowerBound,
      LocalDateTime upperBound,
      BoundType lowerBoundType,
      BoundType upperBoundType) {
    DateBoundary v = new DateBoundary();
    v.lowerBound = lowerBound;
    v.upperBound = upperBound;
    v.lowerBoundType = lowerBoundType;
    v.upperBoundType = upperBoundType;
    v.validate();
    return v;
  }

  @Override
  public LocalDateTime getMinValue() {
    return LocalDateTime.MIN;
  }

  @Override
  public LocalDateTime getMaxValue() {
    return LocalDateTime.MAX;
  }

  @Override
  protected void parse(String text) {
    try {
      text = Validate.notNull(text).trim();
      if (text.isEmpty()) {
        upperBound = getMaxValue();
        lowerBound = getMinValue();
        lowerBoundType = INCLUSIVE;
        upperBoundType = INCLUSIVE;
      } else if (text.startsWith("date")) {
        // Exactly
        Matcher matcher = datePattern.matcher(text);
        if (!matcher.find()) {
          throw new IllegalArgumentException();
        } else {
          String dateTime = matcher.group(dateGroupName);
          lowerBoundType = INCLUSIVE;
          upperBoundType = INCLUSIVE;
          lowerBound = upperBound = LocalDateTime.parse(dateTime);
        }
      } else if (text.startsWith("<")) {
        // Before
        Matcher matcher = datePattern.matcher(text.replace('<', '\0').trim());
        if (!matcher.find()) {
          throw new IllegalArgumentException();
        } else {
          String dateTime = matcher.group(dateGroupName);
          lowerBoundType = INCLUSIVE;
          upperBoundType = EXCLUSIVE;
          lowerBound = getMinValue();
          upperBound = LocalDateTime.parse(dateTime);
        }
      } else if (text.startsWith(">")) {
        // After
        Matcher matcher = datePattern.matcher(text.replace('>', '\0').trim());
        if (!matcher.find()) {
          throw new IllegalArgumentException();
        } else {
          String dateTime = matcher.group(dateGroupName);
          lowerBoundType = EXCLUSIVE;
          upperBoundType = INCLUSIVE;
          lowerBound = LocalDateTime.parse(dateTime);
          upperBound = getMaxValue();
        }
      } else if (text.startsWith("[")) {
        // Between
        String[] textDVals = text.replace('[', '\0').replace(']', '\0').trim().split("\\.\\.");
        Matcher matcher1 = datePattern.matcher(textDVals[0]);
        Matcher matcher2 = datePattern.matcher(textDVals[1]);
        if (!matcher1.find() & !matcher2.find()) {
          throw new IllegalArgumentException();
        } else {
          String dateTime1 = matcher1.group(dateGroupName);
          String dateTime2 = matcher2.group(dateGroupName);
          lowerBoundType = INCLUSIVE;
          upperBoundType = INCLUSIVE;
          lowerBound = LocalDateTime.parse(dateTime1);
          upperBound = LocalDateTime.parse(dateTime2);
        }
      } else {
        throw new IllegalArgumentException();
      }
    } catch (Exception e) {
      throw new BoundaryParseException("Date value " + this.getText() + " couldn't be parsed.");
    }
  }

  @Override
  public String getParsedText() {
    if (lowerBound.equals(getMinValue()) && upperBound.equals(getMaxValue())) {
      return "";
    }
    StringBuilder sb = new StringBuilder();
    if (lowerBound.equals(getMinValue())) {
      sb.append("< ");
      sb.append(CAMUNDA_DATE_TIME_FORMAT.format(upperBound));
    } else if (upperBound.equals(getMaxValue())) {
      sb.append("> ");
      sb.append(CAMUNDA_DATE_TIME_FORMAT.format(lowerBound));
    } else {
      sb.append("[");
      sb.append(CAMUNDA_DATE_TIME_FORMAT.format(lowerBound));
      sb.append("..");
      sb.append(CAMUNDA_DATE_TIME_FORMAT.format(upperBound));
      sb.append("]");
    }
    return sb.toString();
  }
}
