package de.unikoblenz.fgbks.base.utils.boundary.impl;

import static de.unikoblenz.fgbks.base.utils.boundary.BoundType.EXCLUSIVE;
import static de.unikoblenz.fgbks.base.utils.boundary.BoundType.INCLUSIVE;

import de.unikoblenz.fgbks.base.utils.boundary.AbstractGrowingBoundary;
import de.unikoblenz.fgbks.base.utils.boundary.BoundType;
import de.unikoblenz.fgbks.base.utils.boundary.BoundaryParseException;
import org.apache.commons.lang3.Validate;

/**
 * The double boundary is a rage between two double values. And looks as string representation like
 *
 * <ul>
 *   <li>"= 10.0"<br>
 *       {lowerBound=10.0; upperBound=10.0; lowerBoundType={@link BoundType#INCLUSIVE},
 *       upperBoundType={@link BoundType#INCLUSIVE})
 *   <li>"{@literal <}= 10.0"<br>
 *       {lowerBound=-INFINITY; upperBound=10.0; lowerBoundType={@link BoundType#INCLUSIVE},
 *       upperBoundType={@link BoundType#INCLUSIVE})
 *   <li>"{@literal <} 10.0"<br>
 *       {lowerBound=-INFINITY; upperBound=10.0; lowerBoundType={@link BoundType#INCLUSIVE},
 *       upperBoundType={@link BoundType#EXCLUSIVE})
 *   <li>"{@literal >} 10.0"<br>
 *       {lowerBound=10.0; upperBound=INFINITY; lowerBoundType={@link BoundType#EXCLUSIVE},
 *       upperBoundType={@link BoundType#INCLUSIVE})
 *   <li>"{@literal >}= 10.0"<br>
 *       {lowerBound=10.0; upperBound=INFINITY; lowerBoundType={@link BoundType#INCLUSIVE},
 *       upperBoundType={@link BoundType#INCLUSIVE})
 *   <li>"[10.0..20.0]" <br>
 *       {lowerBound=10.0; upperBound=20.0; lowerBoundType={@link BoundType#INCLUSIVE},
 *       upperBoundType={@link BoundType#INCLUSIVE})
 *   <li>"]10.0..20.0]" <br>
 *       {lowerBound=10.0; upperBound=20; lowerBoundType={@link BoundType#EXCLUSIVE},
 *       upperBoundType={@link BoundType#INCLUSIVE})
 *   <li>"]10.0..20.0[" <br>
 *       {lowerBound=10.0; upperBound=20.0; lowerBoundType={@link BoundType#EXCLUSIVE},
 *       upperBoundType={@link BoundType#EXCLUSIVE})
 *   <li>"[10.0..20.0[" <br>
 *       {lowerBound=10.0; upperBound=20.0; lowerBoundType={@link BoundType#INCLUSIVE},
 *       upperBoundType={@link BoundType#EXCLUSIVE})
 *   <li>"" (empty String) <br>
 *       {lowerBound=-INFINITY; upperBound=INFINITY; lowerBoundType={@link BoundType#INCLUSIVE},
 *       upperBoundType={@link BoundType#INCLUSIVE})
 * </ul>
 */
public class DoubleBoundary extends AbstractGrowingBoundary<Double> {

  /**
   * Parse the input text to a new Boolean boundary. See {@link DoubleBoundary} for the required
   * structure.
   *
   * @param text the text, which will be parsed to the new DoubleBoundary object
   */
  public DoubleBoundary(String text) {
    super(text, Double.TYPE);
  }

  private DoubleBoundary() {
    super(Double.TYPE);
  }

  @Override
  public AbstractGrowingBoundary getCopy(
      Double lowerBound, Double upperBound, BoundType lowerBoundType, BoundType upperBoundType) {
    DoubleBoundary v = new DoubleBoundary();
    v.lowerBound = lowerBound;
    v.upperBound = upperBound;
    v.lowerBoundType = lowerBoundType;
    v.upperBoundType = upperBoundType;
    v.validate();
    return v;
  }

  @Override
  public Double getMinValue() {
    return Double.NEGATIVE_INFINITY;
  }

  @Override
  public Double getMaxValue() {
    return Double.POSITIVE_INFINITY;
  }

  @Override
  protected void parse(String text) {
    try {
      text = Validate.notNull(text).trim();
      if (text.isEmpty()) {
        upperBound = getMaxValue();
        lowerBound = getMinValue();
        lowerBoundType = INCLUSIVE;
        upperBoundType = INCLUSIVE;
      } else if (text.startsWith("[") || text.startsWith("]")) {
        lowerBoundType = text.startsWith("[") ? INCLUSIVE : EXCLUSIVE;
        upperBoundType = text.endsWith("]") ? INCLUSIVE : EXCLUSIVE;
        text = text.replace('[', '\0').replace(']', '\0').trim();
        String[] n = text.split("\\.\\.");
        lowerBound = Double.parseDouble(n[0]);
        upperBound = Double.parseDouble(n[1]);
      } else if (text.startsWith(">")) {
        lowerBoundType = text.contains("=") ? INCLUSIVE : EXCLUSIVE;
        text = text.replace('=', '\0').replace('>', '\0').trim();
        lowerBound = Double.parseDouble(text);
        upperBound = getMaxValue();
        upperBoundType = INCLUSIVE;
      } else if (text.startsWith("<")) {
        upperBoundType = text.contains("=") ? INCLUSIVE : EXCLUSIVE;
        text = text.replace('=', '\0').replace('<', '\0').trim();
        upperBound = Double.parseDouble(text);
        lowerBound = getMinValue();
        lowerBoundType = INCLUSIVE;
      } else if (text.startsWith("=")) {
        text = text.replace('=', '\0').trim();
        upperBound = Double.parseDouble(text);
        lowerBound = Double.parseDouble(text);
        lowerBoundType = INCLUSIVE;
        upperBoundType = INCLUSIVE;
      } else {
        throw new BoundaryParseException();
      }
    } catch (Exception e) {
      throw new BoundaryParseException(
          String.format(
              "Parsing Double boundary failed: %s. Exception: %s", this.getText(), e.getMessage()));
    }
  }
}
