package de.unikoblenz.fgbks.base.utils.boundary.impl;

import static de.unikoblenz.fgbks.base.utils.boundary.bicreater.BoundaryBiCreaterType.APPEND;
import static de.unikoblenz.fgbks.base.utils.boundary.bicreater.BoundaryBiCreaterType.BETWEEN;
import static de.unikoblenz.fgbks.base.utils.boundary.bicreater.BoundaryBiCreaterType.COMBINE;
import static de.unikoblenz.fgbks.base.utils.boundary.bicreater.BoundaryBiCreaterType.INTERSECTION;
import static de.unikoblenz.fgbks.base.utils.boundary.bicreater.BoundaryBiCreaterType.LOWER_BOUNDS;
import static de.unikoblenz.fgbks.base.utils.boundary.bicreater.BoundaryBiCreaterType.UPPER_BOUNDS;
import static de.unikoblenz.fgbks.base.utils.boundary.checker.BoundaryCheckType.IS_EQUAL;
import static de.unikoblenz.fgbks.base.utils.boundary.checker.BoundaryCheckType.IS_IN_CONTACT;
import static de.unikoblenz.fgbks.base.utils.boundary.checker.BoundaryCheckType.IS_NOT_IN_CONTACT;
import static de.unikoblenz.fgbks.base.utils.boundary.checker.BoundaryCheckType.IS_OVERLAPPING;
import static de.unikoblenz.fgbks.base.utils.boundary.checker.BoundaryCheckType.IS_SUBSUMPTION;
import static de.unikoblenz.fgbks.base.utils.boundary.checker.BoundaryCheckType.SUBSUMES;
import static de.unikoblenz.fgbks.base.utils.boundary.creater.BoundaryCreaterType.LOWER;
import static de.unikoblenz.fgbks.base.utils.boundary.creater.BoundaryCreaterType.UPPER;

import de.unikoblenz.fgbks.base.builder.DefaultBuilder;
import de.unikoblenz.fgbks.base.utils.boundary.AbstractBoundary;
import de.unikoblenz.fgbks.base.utils.boundary.Boundary;
import de.unikoblenz.fgbks.base.utils.boundary.BoundaryParseException;
import de.unikoblenz.fgbks.base.utils.boundary.bicreater.BiCreaterStringAppend;
import de.unikoblenz.fgbks.base.utils.boundary.bicreater.BiCreaterStringBetween;
import de.unikoblenz.fgbks.base.utils.boundary.bicreater.BiCreaterStringCombine;
import de.unikoblenz.fgbks.base.utils.boundary.bicreater.BiCreaterStringIntersection;
import de.unikoblenz.fgbks.base.utils.boundary.bicreater.BiCreaterStringLowerBounds;
import de.unikoblenz.fgbks.base.utils.boundary.bicreater.BiCreaterStringUpperBounds;
import de.unikoblenz.fgbks.base.utils.boundary.bicreater.BoundaryBiCreater;
import de.unikoblenz.fgbks.base.utils.boundary.bicreater.BoundaryBiCreaterType;
import de.unikoblenz.fgbks.base.utils.boundary.checker.BoundaryCheck;
import de.unikoblenz.fgbks.base.utils.boundary.checker.BoundaryCheckType;
import de.unikoblenz.fgbks.base.utils.boundary.checker.CheckStringEqual;
import de.unikoblenz.fgbks.base.utils.boundary.checker.CheckStringInContact;
import de.unikoblenz.fgbks.base.utils.boundary.checker.CheckStringNotInContact;
import de.unikoblenz.fgbks.base.utils.boundary.checker.CheckStringOverlap;
import de.unikoblenz.fgbks.base.utils.boundary.checker.CheckStringSubsumes;
import de.unikoblenz.fgbks.base.utils.boundary.checker.CheckStringSubsumption;
import de.unikoblenz.fgbks.base.utils.boundary.creater.BoundaryCreater;
import de.unikoblenz.fgbks.base.utils.boundary.creater.BoundaryCreaterType;
import de.unikoblenz.fgbks.base.utils.boundary.creater.CreaterStringLower;
import de.unikoblenz.fgbks.base.utils.boundary.creater.CreaterStringUpper;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Objects;
import java.util.Set;
import java.util.stream.Collectors;
import org.apache.commons.lang3.Validate;

/**
 * String Boundary, representing a set of strings. If the boundary contains all possible string, the
 * method {@link StringBoundary#matchesAny()} returns true. The method {@link
 * StringBoundary#getValues()} returns all string, which are in the boundary, or if {@link
 * StringBoundary#matchesNoneOfValues} returns true, which are not in the boundary. <br>
 *
 * <p>Only value "a": <br>
 * {@code "a"} <br>
 * Value "a" or value "b": <br>
 * {@code "a","b"} <br>
 * Not value "a" and not value "b": <br>
 * {@code not("a","b")} <br>
 * All values: <br>
 * {@code (empty String)}
 */
public class StringBoundary extends AbstractBoundary<String> {

  private static HashMap<BoundaryCheckType, BoundaryCheck> checkerMap = new HashMap<>();
  private static HashMap<BoundaryCreaterType, BoundaryCreater> createrMap = new HashMap<>();
  private static HashMap<BoundaryBiCreaterType, BoundaryBiCreater> biCreaterMap = new HashMap<>();

  static {
    checkerMap.put(IS_EQUAL, CheckStringEqual.getInstance());
    checkerMap.put(IS_IN_CONTACT, CheckStringInContact.getInstance());
    checkerMap.put(IS_NOT_IN_CONTACT, CheckStringNotInContact.getInstance());
    checkerMap.put(IS_OVERLAPPING, CheckStringOverlap.getInstance());
    checkerMap.put(IS_SUBSUMPTION, CheckStringSubsumption.getInstance());
    checkerMap.put(SUBSUMES, CheckStringSubsumes.getInstance());
    createrMap.put(UPPER, CreaterStringUpper.getInstance());
    createrMap.put(LOWER, CreaterStringLower.getInstance());
    biCreaterMap.put(INTERSECTION, BiCreaterStringIntersection.getInstance());
    biCreaterMap.put(LOWER_BOUNDS, BiCreaterStringLowerBounds.getInstance());
    biCreaterMap.put(UPPER_BOUNDS, BiCreaterStringUpperBounds.getInstance());
    biCreaterMap.put(BETWEEN, BiCreaterStringBetween.getInstance());
    biCreaterMap.put(APPEND, BiCreaterStringAppend.getInstance());
    biCreaterMap.put(COMBINE, BiCreaterStringCombine.getInstance());
  }

  private boolean matchesNoneOfValues;
  private boolean matchesAny;
  private String[] values;
  // sorted list of cached hashes of valuesHashes
  private int[] valuesHashes;

  /**
   * Parse the input text to a new String boundary. See {@link StringBoundary} for the required
   * structure or use {@link StringBoundary#getBuilder()} to create a StringBoundary with a builder.
   *
   * @param text the text, which will be parsed to the new StringBoundary object
   */
  public StringBoundary(String text) {
    super(text, String.class);
  }

  private StringBoundary() {
    super(String.class);
  }

  @Override
  public int compareTo(Boundary o) {
    if (!(o instanceof StringBoundary)) {
      return 0;
    }
    StringBoundary otherStringB = (StringBoundary) o;
    if (matchesAny() || matchesNoneOfValues()) {
      if (otherStringB.matchesAny() || otherStringB.matchesNoneOfValues()) {
        return 0;
      } else {
        return -1;
      }
    }
    if (otherStringB.matchesAny() || otherStringB.matchesNoneOfValues()) {
      return 1;
    }
    return this.getValues()[0].compareTo(otherStringB.getValues()[0]);
  }

  /**
   * Get a new builder object of the StringBoundary.
   *
   * @return a new builder
   */
  public static Builder getBuilder() {
    return new StringBoundary().new Builder();
  }

  /**
   * Get the boolean, if the StringBoundary matches none of the values.
   *
   * @return true, if the StringBoundary matches none of the values
   */
  public boolean matchesNoneOfValues() {
    return matchesNoneOfValues;
  }

  /**
   * Get the boolean, if the StringBoundary matches all possible values.
   *
   * @return true, if the StringBoundary matches all possible values
   */
  public boolean matchesAny() {
    return matchesAny;
  }

  /**
   * Get all the values, which are representing the String boundary. The created array is a copy of
   * the original one.
   *
   * @return all values as String array
   */
  public String[] getValues() {
    return values.clone(); // performance ?
  }

  /**
   * Get all the hashes of the string values as sorted array. The created array is a copy of the
   * original one.
   *
   * @return all hashes of the values as int array
   */
  public int[] getValuesHashes() {
    return valuesHashes.clone(); // performance ?
  }

  @Override
  protected void parse(String text) {
    matchesNoneOfValues = false;
    matchesAny = false;
    try {
      matchesAny = Validate.notNull(text).isEmpty();
      if (matchesAny) {
        values = new String[0];
      } else {
        matchesNoneOfValues = text.toLowerCase().startsWith("not");
        text = text.replace('(', '\0').replace(')', '\0').trim();
        if (matchesNoneOfValues) {
          text = text.replaceFirst("not", "").trim();
        }
        values = text.split(",");
      }
      calcHashes();
    } catch (Exception e) {
      throw new BoundaryParseException("String value is not valid: " + this.getText());
    }
  }

  private void calcHashes() {
    Arrays.sort(values);
    valuesHashes = new int[values.length];
    for (int i = 0; i < values.length; i++) {
      values[i] = values[i].replace('"', '\0').trim();
      valuesHashes[i] = Objects.hashCode(values[i]);
    }
    // Sort hashes for later comparisons
    Arrays.sort(valuesHashes);
  }

  /**
   * Get the amount of string values. This method is preferred to get the size of the result from
   * the {@link StringBoundary#getValues()} method, because {@link StringBoundary#getValues()}
   * creates a copy of the array.
   *
   * @return the size of the array from {@link StringBoundary#getValues()}.
   */
  public int getAmountOfElements() {
    return values.length;
  }

  @Override
  protected Map<BoundaryCheckType, BoundaryCheck> getChecker() {
    return Collections.unmodifiableMap(checkerMap);
  }

  @Override
  protected Map<BoundaryBiCreaterType, BoundaryBiCreater> getBiCreater() {
    return Collections.unmodifiableMap(biCreaterMap);
  }

  @Override
  protected Map<BoundaryCreaterType, BoundaryCreater> getCreater() {
    return Collections.unmodifiableMap(createrMap);
  }

  @Override
  protected void validate() {
    try {
      Validate.noNullElements(values);
    } catch (Exception e) {
      throw new BoundaryParseException(
          "String value " + getText() + " could not be parsed to a StringBoundary");
    }
  }

  /**
   * {@inheritDoc}
   */
  @Override
  public String getParsedText() {
    return (matchesNoneOfValues ? "not(" : "")
        + Arrays.stream(values).map(v -> '"' + v + '"').collect(Collectors.joining(","))
        + (matchesNoneOfValues ? ")" : "");
  }

  /** Builder to create new {@link StringBoundary}s. */
  public class Builder extends DefaultBuilder<StringBoundary> {

    private Set<String> buildingValues;

    private Builder() {
      buildingValues = new HashSet<>();
      value.matchesAny = false;
      value.matchesNoneOfValues = false;
    }

    /** See {@link StringBoundary#matchesNoneOfValues()}. */
    public Builder matchesNoneOfValues(boolean matchesNoneOfValues) {
      value.matchesNoneOfValues = matchesNoneOfValues;
      return this;
    }

    /** See {@link StringBoundary#matchesAny()}. */
    public Builder matchesAny(boolean matchAny) {
      value.matchesAny = matchAny;
      return this;
    }

    /**
     * Add a value to the StringBoundary.
     *
     * @param value the String value
     * @return the current builder
     */
    public Builder addValue(String value) {
      buildingValues.add(Validate.notNull(value));
      return this;
    }

    /**
     * Add values to the StringBoundary.
     *
     * @param values the String values
     * @return the current builder
     */
    public Builder addValues(Collection<String> values) {
      buildingValues.addAll(Validate.noNullElements(values));
      return this;
    }

    /**
     * Add values to the StringBoundary.
     *
     * @param values the String values
     * @return the current builder
     */
    public Builder addValues(String[] values) {
      return addValues(Validate.notNull(Arrays.asList(values)));
    }

    @Override
    protected void validate() {
      super.validate();
      value.values = buildingValues.toArray(new String[0]);
      if (value.matchesAny) {
        value.matchesNoneOfValues = false;
        value.values = new String[0];
      }
      value.calcHashes();
      value.text = value.getParsedText();
      value.validate();
    }

    /**
     * Build the StringBoundary.
     *
     * @return a new instance of the {@link StringBoundary}
     */
    @Override
    public StringBoundary build() {
      return super.build();
    }
  }
}
