package de.unikoblenz.fgbks.base.value;

import static java.util.Objects.nonNull;
import static org.apache.commons.lang3.ArrayUtils.subarray;

import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.math.BigDecimal;
import java.math.RoundingMode;

/**
 * Value Object
 */
public abstract class AbstractBigDecimalValueObject<T extends AbstractBigDecimalValueObject>
    extends AbstractSimpleValueObject<BigDecimal> {
  private static final BigDecimal DEFAULT_MAX_VALUE = new BigDecimal(Long.MAX_VALUE);

  protected AbstractBigDecimalValueObject() {}

  public AbstractBigDecimalValueObject(BigDecimal initialValue) {
    super(initialValue);
  }

  public AbstractBigDecimalValueObject(String initialValue) {
    super(new BigDecimal(initialValue));
  }

  @Override
  public BigDecimal getValue() {
    return normalize(super.getValue());
  }

  protected BigDecimal getMinValue() {
    return DEFAULT_MAX_VALUE.negate();
  }

  protected BigDecimal getMaxValue() {
    return DEFAULT_MAX_VALUE;
  }

  @Override
  protected BigDecimal validateAndNormalize(BigDecimal newValue) {
    BigDecimal retValue = super.validateAndNormalize(newValue);
    if (getMinValue().compareTo(retValue) > 0) {
      throw new OverflowException("Minimum value is " + getMinValue().toString());
    }
    if (getMaxValue().compareTo(retValue) < 0) {
      throw new OverflowException("Maximum value is " + getMaxValue().toString());
    }
    return normalize(retValue);
  }

  protected int getScale() {
    return -1;
  }

  protected BigDecimal normalize(BigDecimal aValue) {
    if (getScale() == -1) {
      return aValue;
    } else {
      return aValue.setScale(getScale(), RoundingMode.HALF_UP);
    }
  }

  public T add(T value) {
    return add(value.getValue());
  }

  public T add(BigDecimal value) {
    return newInstance(getValue().add(value));
  }

  public T subtract(T value) {
    return subtract(value.getValue());
  }

  public T subtract(BigDecimal value) {
    return newInstance(getValue().subtract(value));
  }

  public T multiply(T value) {
    return multiply(value.getValue());
  }

  public T multiply(BigDecimal value) {
    return newInstance(getValue().multiply(value));
  }

  public T multiply(long value) {
    return newInstance(getValue().multiply(BigDecimal.valueOf(value)));
  }

  public T divide(T value) {
    return divide(value.getValue());
  }

  public T divide(BigDecimal value) {
    try {
      return newInstance(getValue().divide(value, getScale(), RoundingMode.HALF_UP));
    } catch (ArithmeticException e) {
      throw new IllegalStateException(e.getMessage());
    }
  }

  public T negate() {
    return newInstance(getValue().negate());
  }

  public T abs() {
    return newInstance(getValue().abs());
  }

  public boolean greaterThan(T value) {
    return getValue().compareTo(value.getValue()) > 0;
  }

  public boolean greaterThan(BigDecimal value) {
    return getValue().compareTo(value) > 0;
  }

  public boolean greaterEqualThan(T value) {
    return getValue().compareTo(value.getValue()) >= 0;
  }

  public boolean greaterEqualThan(BigDecimal value) {
    return getValue().compareTo(value) >= 0;
  }

  public boolean lessEqualThan(T value) {
    return getValue().compareTo(value.getValue()) <= 0;
  }

  public boolean lessEqualThan(BigDecimal value) {
    return getValue().compareTo(value) <= 0;
  }

  public boolean lessThan(T value) {
    return getValue().compareTo(value.getValue()) < 0;
  }

  public boolean lessThan(BigDecimal value) {
    return getValue().compareTo(value) < 0;
  }

  public boolean equalTo(T value) {
    return getValue().compareTo(value.getValue()) == 0;
  }

  public boolean equalTo(BigDecimal value) {
    return getValue().compareTo(value) == 0;
  }

  protected BigDecimal trimTrailingZerosFrom(BigDecimal bigDecimal) {
    char[] c = bigDecimal.toString().toCharArray();
    c = subarray(c, c.length - bigDecimal.scale(), c.length);
    for (int i = bigDecimal.scale() - 1; i >= 0; i--) {
      if (c[i] != '0') {
        return bigDecimal.setScale(i + 1);
      }
    }
    return bigDecimal.setScale(0);
  }

  protected T newInstance(BigDecimal initialValue) {
    try {
      Type genericSuperclass = this.getClass().getGenericSuperclass();
      Class aClass = (Class) ((ParameterizedType) genericSuperclass).getActualTypeArguments()[0];
      return (T) aClass.getDeclaredConstructor(BigDecimal.class).newInstance(initialValue);
    } catch (Exception e) {
      if (nonNull(e.getCause())) {
        throw new IllegalStateException(e.getCause());
      }
      throw new IllegalStateException(e);
    }
  }
}
