package de.unikoblenz.fgbks.base.value;

import static java.util.Objects.nonNull;
import static org.apache.commons.lang3.Validate.notNull;

import java.io.Serializable;
import javax.xml.bind.annotation.XmlTransient;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

@XmlTransient
public abstract class AbstractSimpleValueObject<V extends Comparable<V>>
    implements Serializable, Comparable<AbstractSimpleValueObject<V>>, SimpleValueObject {

  public static final String VALUE = "value";
  private static final Logger LOG = LoggerFactory.getLogger(AbstractSimpleValueObject.class);

  private V value;

  protected AbstractSimpleValueObject() {
    // required for proxying
  }

  public AbstractSimpleValueObject(V initialValue) {
    value = validateAndNormalize(initialValue);
  }

  /**
   * Validates and normalizes the constructor parameter of this simple value object. This
   * implementation checks the value to be non-null. Subclasses may override this method to alter
   * validation and normalization.
   *
   * @param newValue The constructor value.
   * @return The validated and normalized value.
   */
  protected V validateAndNormalize(V newValue) {
    return notNull(newValue, "value may not be null");
  }

  public V getValue() {
    return value;
  }

  @Override
  public String toString() {
    return getValue().toString();
  }

  @Override
  public int hashCode() {
    return getValue().hashCode();
  }

  @Override
  public boolean equals(Object object) {
    if (this == object) {
      return true;
    }
    if (object == null || !getClass().isAssignableFrom(object.getClass())) {
      return false;
    }
    AbstractSimpleValueObject<V> valueObject = (AbstractSimpleValueObject<V>) object;
    return valueObject.getValue().equals(getValue());
  }

  @Override
  public int compareTo(AbstractSimpleValueObject<V> object) {
    return getValue().compareTo(object.getValue());
  }

  public boolean nonNullValue() {
    return nonNull(getValue());
  }
}
