package de.unikoblenz.fgbks.base.wordnet;

import edu.mit.jwi.Dictionary;
import edu.mit.jwi.IDictionary;
import edu.mit.jwi.item.IIndexWord;
import edu.mit.jwi.item.ISynset;
import edu.mit.jwi.item.IWord;
import edu.mit.jwi.item.IWordID;
import edu.mit.jwi.item.POS;
import java.io.File;
import java.net.URL;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

/**
 * WordnetService provides some functions from wordnet <br>
 * First, get the instance with getInstance() (singleton) and than use the provided functions.
 */
public final class WordnetService implements WordnetServiceFunctions {

  private static final String PATH_WORDNET = "dict/";
  // Singleton pattern
  private static WordnetService instance;

  private IDictionary dictionary;

  private WordnetService() {
    setupDictionary();
  }

  /**
   * Get the singleton instance of the wordnet service.
   *
   * @param refresh true for a new initialisation of wordnet
   * @return the wordnet service instance
   */
  public static WordnetService getInstance(boolean refresh) {
    if (instance == null) {
      instance = new WordnetService();
    } else if (refresh) {
      instance.setupDictionary();
    }
    return instance;
  }

  /**
   * Get the singleton instance of the wordnet service.
   *
   * @return the wordnet service instance
   */
  public static WordnetService getInstance() {
    return getInstance(false);
  }

  private void setupDictionary() {
    if (dictionary != null) {
      dictionary.close();
    }
    dictionary = null;
    try {
      String totalPath =
          new File(getClass().getClassLoader().getResource(PATH_WORDNET).getFile())
              .getAbsolutePath();
      dictionary = new Dictionary(new URL("file", null, totalPath));
      dictionary.open();
    } catch (Exception e) {
      dictionary = null;
    }
  }

  /**
   * Return true, if both nouns are in one synset. <br>
   * "examination" {@literal &} "exam" = true <br>
   * "bill" {@literal &} "invoice" = true<br>
   *
   * @return true, if both nouns are synonyms
   */
  @Override
  public boolean areNounsSynonyms(String noun1, String noun2) {
    if (dictionary == null) {
      return false;
    }
    noun1 = noun1.toLowerCase().replace(" ", "_");
    noun2 = noun2.toLowerCase().replace(" ", "_");
    if (noun1.equals(noun2)) {
      return false;
    }
    return getNounsSynonyms(noun1).contains(noun2);
  }

  @Override
  public Set<String> getNounsSynonyms(String noun) {
    if (dictionary == null) {
      return Collections.emptySet();
    }
    noun = noun.toLowerCase().replace(" ", "_");
    IIndexWord idxWord = null;
    try {
      idxWord = dictionary.getIndexWord(noun, POS.NOUN);
    } catch (Exception e) {
      // nothing
    }
    if (idxWord == null) {
      return Collections.emptySet();
    }
    List<IWordID> wordIDs = idxWord.getWordIDs();
    Set<String> returnSet = new HashSet<>();
    for (IWordID wordID : wordIDs) {
      IWord word = dictionary.getWord(wordID);
      ISynset synset = word.getSynset();

      // iterate over words associated with the synset
      for (IWord w : synset.getWords()) {
        returnSet.add(w.getLemma());
      }
    }
    return returnSet;
  }

  /** Close the wordnet dictionary. */
  public void close() {
    if (dictionary != null && dictionary.isOpen()) {
      dictionary.close();
    }
  }
}
